
ADMISSION_TIME_COL = 'admittime'
DISCHARGE_TIME_COL = 'dischtime'
DEATH_TIME_COL = 'deathtime'
ED_REG_TIME = 'edregtime'
ED_OUT_TIME = 'edouttime'
AGE_COL = 'anchor_age'
GENDER_COL = 'gender'

AGE_BINS = list(range(0, 125, 5))
AGE_LABELS = [f'{AGE_BINS[a]}' for a in range(len(AGE_BINS) - 1)]

font_sz = 14
title_sz = 18

YEAR_GROUP_COL = 'anchor_year_group'
SUBSET_YEAR_GROUP = '2017 - 2019'
SUBJECT_ID_COL = 'subject_id'
ADMISSION_ID_COL = 'hadm_id'
ADMISSION_TYPE_COL = 'admission_type'
CHART_TIME_COL = 'charttime'
STORE_TIME_COL = 'storetime'
LOS_EXACT_COL = 'LOS exact'
LOS_DAYS_COL = 'LOS days'
ADMISSION_LOCATION_COL = 'admission_location'
DISCHARGE_LOCATION_COL = 'discharge_location'
RACE_COL = 'race'
INSURANCE_COL = 'insurance'
ADMISSION_TO_RESULT_COL = 'admission_to_result_time'
ADMISSION_AGE_COL = 'admission_age'
ADMISSION_YEAR_COL = 'admission_year'
ADMISSION_COUNT_COL = 'admissions_count'
ITEM_ID_COL = 'itemid'
NIGHT_ADMISSION_FLAG = 'night_admission'
MARITAL_STATUS_COL = 'marital_status'
STANDARDIZED_AGE_COL = 'standardized_age'
COEF_COL = '   coef   '
STDERR_COL = ' std err '
DIRECT_IND_COL = 'direct_emrgency_flag'
PREV_ADMISSION_IND_COL = 'last_less_than_diff'
ADMISSION_COUNT_GROUP_COL = ADMISSION_COUNT_COL + '_group'

DISCHARGE_REGROUPING_DICT = {
    'HOME': 'HOME',
    'HOME HEALTH CARE': 'HOME',
    'SKILLED NURSING FACILITY': 'FURTHER TREATMENT',
    'DIED': 'DIED',
    'REHAB': 'HOME',
    'CHRONIC/LONG TERM ACUTE CARE': 'FURTHER TREATMENT',
    'HOSPICE': 'FURTHER TREATMENT',
    'AGAINST ADVICE': 'CENSORED',
    'ACUTE HOSPITAL': 'FURTHER TREATMENT',
    'PSYCH FACILITY': 'FURTHER TREATMENT',
    'OTHER FACILITY': 'FURTHER TREATMENT',
    'ASSISTED LIVING': 'HOME',
    'HEALTHCARE FACILITY': 'FURTHER TREATMENT',
}

RACE_REGROUPING_DICT = {
    'WHITE': 'WHITE',
    'UNKNOWN': 'OTHER',
    'BLACK/AFRICAN AMERICAN': 'BLACK',
    'OTHER': 'OTHER',
    'ASIAN': 'ASIAN',
    'WHITE - OTHER EUROPEAN': 'WHITE',
    'HISPANIC/LATINO - PUERTO RICAN': 'HISPANIC',
    'HISPANIC/LATINO - DOMINICAN': 'HISPANIC',
    'ASIAN - CHINESE': 'ASIAN',
    'BLACK/CARIBBEAN ISLAND': 'BLACK',
    'BLACK/AFRICAN': 'BLACK',
    'BLACK/CAPE VERDEAN': 'BLACK',
    'PATIENT DECLINED TO ANSWER': 'OTHER',
    'WHITE - BRAZILIAN': 'WHITE',
    'PORTUGUESE': 'HISPANIC',
    'ASIAN - SOUTH EAST ASIAN': 'ASIAN',
    'WHITE - RUSSIAN': 'WHITE',
    'ASIAN - ASIAN INDIAN': 'ASIAN',
    'WHITE - EASTERN EUROPEAN': 'WHITE',
    'AMERICAN INDIAN/ALASKA NATIVE': 'OTHER',
    'HISPANIC/LATINO - GUATEMALAN': 'HISPANIC',
    'HISPANIC/LATINO - MEXICAN': 'HISPANIC',
    'HISPANIC/LATINO - SALVADORAN': 'HISPANIC',
    'SOUTH AMERICAN': 'HISPANIC',
    'NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER': 'OTHER',
    'HISPANIC/LATINO - COLUMBIAN': 'HISPANIC',
    'HISPANIC/LATINO - CUBAN': 'HISPANIC',
    'ASIAN - KOREAN': 'ASIAN',
    'HISPANIC/LATINO - HONDURAN': 'HISPANIC',
    'HISPANIC/LATINO - CENTRAL AMERICAN': 'HISPANIC',
    'UNABLE TO OBTAIN': 'OTHER',
    'HISPANIC OR LATINO': 'HISPANIC'
}

# 'MCH': 'Mean Cell Hemoglobin',
# 'MCHC': 'Mean Cell Hemoglobin Concentration',
table1_rename_columns = {
    'AnionGap': 'Anion gap',
    'Bicarbonate': 'Bicarbonate',
    'CalciumTotal': 'Calcium total',
    'Chloride': 'Chloride',
    'Creatinine': 'Creatinine',
    'Glucose': 'Glucose',
    'Magnesium': 'Magnesium',
    'Phosphate': 'Phosphate',
    'Potassium': 'Potassium',
    'Sodium': 'Sodium',
    'UreaNitrogen': 'Urea nitrogen',
    'Hematocrit': 'Hematocrit',
    'Hemoglobin': 'Hemoglobin',
    'MCH': 'MCH',
    'MCHC': 'MCHC',
    'MCV': 'MCV',
    'PlateletCount': 'Platelet count',
    'RDW': 'RDW',
    'RedBloodCells': 'Red blood cells',
    'WhiteBloodCells': 'White blood cells',
    NIGHT_ADMISSION_FLAG: 'Night admission',
    GENDER_COL: 'Sex',
    DIRECT_IND_COL: 'Direct emergency',
    PREV_ADMISSION_IND_COL: 'Previous admission this month',
    ADMISSION_AGE_COL: 'Admission age',
    INSURANCE_COL: 'Insurance',
    MARITAL_STATUS_COL: 'Marital status',
    RACE_COL: 'Race',
    ADMISSION_COUNT_GROUP_COL: 'Admissions number',
    LOS_DAYS_COL: 'LOS (days)',
    DISCHARGE_LOCATION_COL: 'Discharge location'
}

table1_rename_sex = {0: 'Male', 1: 'Female'}
table1_rename_race = {'ASIAN': 'Asian', 'BLACK': 'Black', 'HISPANIC': 'Hispanic', 'OTHER': 'Other',
                      'WHITE': 'White'}
table1_rename_marital = {'SINGLE': 'Single', 'MARRIED': 'Married', 'DIVORCED': 'Divorced', 'WIDOWED': 'Widowed'}
table1_rename_yes_no = {0: 'No', 1: 'Yes'}
table1_rename_normal_abnormal = {0: 'Normal', 1: 'Abnormal'}
table1_rename_discharge = {1: 'Home', 2: 'Further Treatment', 3: 'Died', 0: 'Censored'}




rename_beta_index = {
    'AdmsCount 2': 'Admissions number 2',
    'AdmsCount 3up': 'Admissions number 3+',
    'AnionGap': 'Anion gap',
    'Bicarbonate': 'Bicarbonate',
    'CalciumTotal': 'Calcium total',
    'Chloride': 'Chloride',
    'Creatinine': 'Creatinine',
    'Ethnicity BLACK': 'Ethnicity black',
    'Ethnicity HISPANIC': 'Ethnicity hispanic',
    'Ethnicity OTHER': 'Ethnicity other',
    'Ethnicity WHITE': 'Ethnicity white',
    'Glucose': 'Glucose',
    'Hematocrit': 'Hematocrit',
    'Hemoglobin': 'Hemoglobin',
    'Insurance Medicare': 'Insurance medicare',
    'Insurance Other': 'Insurance other',
    'MCH': 'MCH',
    'MCHC': 'MCHC',
    'MCV': 'MCV',
    'Magnesium': 'Magnesium',
    'Marital MARRIED': 'Marital married',
    'Marital SINGLE': 'Marital single',
    'Marital WIDOWED': 'Marital widowed',
    'Phosphate': 'Phosphate',
    'PlateletCount': 'Platelet count',
    'Potassium': 'Potassium',
    'RDW': 'RDW',
    'RedBloodCells': 'Red blood cells',
    'Sodium': 'Sodium',
    'UreaNitrogen': 'Urea nitrogen',
    'WhiteBloodCells': 'White blood cells',
    'direct emrgency flag': 'Direct emergency',
    'gender': 'Sex',
    'last less than diff': 'Recent admission',
    'night admission': 'Night admission',
    'standardized age': 'Standardized age',
}

beta_units = {
    'Admissions number 2': '2',
    'Admissions number 3+': '3+',
    'Anion gap': 'Abnormal',
    'Bicarbonate': 'Abnormal',
    'Calcium total': 'Abnormal',
    'Chloride': 'Abnormal',
    'Creatinine': 'Abnormal',
    'Ethnicity black': 'Black',
    'Ethnicity hispanic': 'Hispanic',
    'Ethnicity other': 'Other',
    'Ethnicity white': 'White',
    'Glucose': 'Abnormal',
    'Hematocrit': 'Abnormal',
    'Hemoglobin': 'Abnormal',
    'Insurance medicare': 'Medicare',
    'Insurance other': 'Other',
    'MCH': 'Abnormal',
    'MCHC': 'Abnormal',
    'MCV': 'Abnormal',
    'Magnesium': 'Abnormal',
    'Marital married': 'Married',
    'Marital single': 'Single',
    'Marital widowed': 'Widowed',
    'Phosphate': 'Abnormal',
    'Platelet count': 'Abnormal',
    'Potassium': 'Abnormal',
    'RDW': 'Abnormal',
    'Red blood cells': 'Abnormal',
    'Sodium': 'Abnormal',
    'Urea nitrogen': 'Abnormal',
    'White blood cells': 'Abnormal',
    'Direct emergency': 'Yes',
    'Sex': 'Female',
    'Recent admission': 'Yes',
    'Night admission': 'Yes',
    'Standardized age': '',
}

# ADMISSION_TIME_COL = 'admittime'
# DISCHARGE_TIME_COL = 'dischtime'
# DEATH_TIME_COL = 'deathtime'
# ED_REG_TIME = 'edregtime'
# ED_OUT_TIME = 'edouttime'
# AGE_COL = 'anchor_age'
# GENDER_COL = 'gender'
#
# AGE_BINS = list(range(0, 125, 5))
# AGE_LABELS = [f'{AGE_BINS[a]}' for a in range(len(AGE_BINS) - 1)]
#
# font_sz = 14
# title_sz = 18
#
# YEAR_GROUP_COL = 'anchor_year_group'
# SUBSET_YEAR_GROUP = '2017 - 2019'
# SUBJECT_ID_COL = 'subject_id'
# ADMISSION_ID_COL = 'hadm_id'
# ADMISSION_TYPE_COL = 'admission_type'
# CHART_TIME_COL = 'charttime'
# STORE_TIME_COL = 'storetime'
# LOS_EXACT_COL = 'LOS exact'
# LOS_DAYS_COL = 'LOS days'
# ADMISSION_LOCATION_COL = 'admission_location'
# DISCHARGE_LOCATION_COL = 'discharge_location'
# RACE_COL = 'race'
# INSURANCE_COL = 'insurance'
# ADMISSION_TO_RESULT_COL = 'admission_to_result_time'
# ADMISSION_AGE_COL = 'admission_age'
# ADMISSION_YEAR_COL = 'admission_year'
# ADMISSION_COUNT_COL = 'admissions_count'
# ITEM_ID_COL = 'itemid'
# NIGHT_ADMISSION_FLAG = 'night_admission'
# MARITAL_STATUS_COL = 'marital_status'
# STANDARDIZED_AGE_COL = 'standardized_age'
# COEF_COL = '   coef   '
# STDERR_COL = ' std err '
# DIRECT_IND_COL = 'direct_emrgency_flag'
# PREV_ADMISSION_IND_COL = 'last_less_than_diff'
# ADMISSION_COUNT_GROUP_COL = ADMISSION_COUNT_COL + '_group'
#
# DISCHARGE_REGROUPING_DICT = {
#     'HOME': 'HOME',
#     'HOME HEALTH CARE': 'HOME',
#     'SKILLED NURSING FACILITY': 'FURTHER TREATMENT',
#     'DIED': 'DIED',
#     'REHAB': 'HOME',
#     'CHRONIC/LONG TERM ACUTE CARE': 'FURTHER TREATMENT',
#     'HOSPICE': 'FURTHER TREATMENT',
#     'AGAINST ADVICE': 'CENSORED',
#     'ACUTE HOSPITAL': 'FURTHER TREATMENT',
#     'PSYCH FACILITY': 'FURTHER TREATMENT',
#     'OTHER FACILITY': 'FURTHER TREATMENT',
#     'ASSISTED LIVING': 'HOME',
#     'HEALTHCARE FACILITY': 'FURTHER TREATMENT',
# }
#
# RACE_REGROUPING_DICT = {
#     'WHITE': 'WHITE',
#     'UNKNOWN': 'OTHER',
#     'BLACK/AFRICAN AMERICAN': 'BLACK',
#     'OTHER': 'OTHER',
#     'ASIAN': 'ASIAN',
#     'WHITE - OTHER EUROPEAN': 'WHITE',
#     'HISPANIC/LATINO - PUERTO RICAN': 'HISPANIC',
#     'HISPANIC/LATINO - DOMINICAN': 'HISPANIC',
#     'ASIAN - CHINESE': 'ASIAN',
#     'BLACK/CARIBBEAN ISLAND': 'BLACK',
#     'BLACK/AFRICAN': 'BLACK',
#     'BLACK/CAPE VERDEAN': 'BLACK',
#     'PATIENT DECLINED TO ANSWER': 'OTHER',
#     'WHITE - BRAZILIAN': 'WHITE',
#     'PORTUGUESE': 'HISPANIC',
#     'ASIAN - SOUTH EAST ASIAN': 'ASIAN',
#     'WHITE - RUSSIAN': 'WHITE',
#     'ASIAN - ASIAN INDIAN': 'ASIAN',
#     'WHITE - EASTERN EUROPEAN': 'WHITE',
#     'AMERICAN INDIAN/ALASKA NATIVE': 'OTHER',
#     'HISPANIC/LATINO - GUATEMALAN': 'HISPANIC',
#     'HISPANIC/LATINO - MEXICAN': 'HISPANIC',
#     'HISPANIC/LATINO - SALVADORAN': 'HISPANIC',
#     'SOUTH AMERICAN': 'HISPANIC',
#     'NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER': 'OTHER',
#     'HISPANIC/LATINO - COLUMBIAN': 'HISPANIC',
#     'HISPANIC/LATINO - CUBAN': 'HISPANIC',
#     'ASIAN - KOREAN': 'ASIAN',
#     'HISPANIC/LATINO - HONDURAN': 'HISPANIC',
#     'HISPANIC/LATINO - CENTRAL AMERICAN': 'HISPANIC',
#     'UNABLE TO OBTAIN': 'OTHER',
#     'HISPANIC OR LATINO': 'HISPANIC'
# }
#
# # 'MCH': 'Mean Cell Hemoglobin',
# # 'MCHC': 'Mean Cell Hemoglobin Concentration',
# table1_rename_columns = {
#     'AnionGap': 'Anion gap',
#     'Bicarbonate': 'Bicarbonate',
#     'CalciumTotal': 'Calcium total',
#     'Chloride': 'Chloride',
#     'Creatinine': 'Creatinine',
#     'Glucose': 'Glucose',
#     'Magnesium': 'Magnesium',
#     'Phosphate': 'Phosphate',
#     'Potassium': 'Potassium',
#     'Sodium': 'Sodium',
#     'UreaNitrogen': 'Urea nitrogen',
#     'Hematocrit': 'Hematocrit',
#     'Hemoglobin': 'Hemoglobin',
#     'MCH': 'MCH',
#     'MCHC': 'MCHC',
#     'MCV': 'MCV',
#     'PlateletCount': 'Platelet count',
#     'RDW': 'RDW',
#     'RedBloodCells': 'Red blood cells',
#     'WhiteBloodCells': 'White blood cells',
#     NIGHT_ADMISSION_FLAG: 'Night admission',
#     GENDER_COL: 'Sex',
#     DIRECT_IND_COL: 'Direct emergency',
#     PREV_ADMISSION_IND_COL: 'Previous admission this month',
#     ADMISSION_AGE_COL: 'Admission age',
#     INSURANCE_COL: 'Insurance',
#     MARITAL_STATUS_COL: 'Marital status',
#     RACE_COL: 'Race',
#     ADMISSION_COUNT_GROUP_COL: 'Admissions number',
#     LOS_DAYS_COL: 'LOS (days)',
#     DISCHARGE_LOCATION_COL: 'Discharge location'
# }
#
# table1_rename_sex = {0: 'Male', 1: 'Female'}
# table1_rename_race = {'ASIAN': 'Asian', 'BLACK': 'Black', 'HISPANIC': 'Hispanic', 'OTHER': 'Other',
#                       'WHITE': 'White'}
# table1_rename_marital = {'SINGLE': 'Single', 'MARRIED': 'Married', 'DIVORCED': 'Divorced', 'WIDOWED': 'Widowed'}
# table1_rename_yes_no = {0: 'No', 1: 'Yes'}
# table1_rename_normal_abnormal = {0: 'Normal', 1: 'Abnormal'}
# table1_rename_discharge = {1: 'Home', 2: 'Further Treatment', 3: 'Died', 0: 'Censored'}

# rename_beta_index = {
#     'AdmsCount 2': 'Admissions number 2',
#     'AdmsCount 3up': 'Admissions number 3+',
#     'AnionGap': 'Anion gap',
#     'Bicarbonate': 'Bicarbonate',
#     'CalciumTotal': 'Calcium total',
#     'Chloride': 'Chloride',
#     'Creatinine': 'Creatinine',
#     'Ethnicity BLACK': 'Ethnicity black',
#     'Ethnicity HISPANIC': 'Ethnicity hispanic',
#     'Ethnicity OTHER': 'Ethnicity other',
#     'Ethnicity WHITE': 'Ethnicity white',
#     'Glucose': 'Glucose',
#     'Hematocrit': 'Hematocrit',
#     'Hemoglobin': 'Hemoglobin',
#     'Insurance Medicare': 'Insurance medicare',
#     'Insurance Other': 'Insurance other',
#     'MCH': 'MCH',
#     'MCHC': 'MCHC',
#     'MCV': 'MCV',
#     'Magnesium': 'Magnesium',
#     'Marital MARRIED': 'Marital married',
#     'Marital SINGLE': 'Marital single',
#     'Marital WIDOWED': 'Marital widowed',
#     'Phosphate': 'Phosphate',
#     'PlateletCount': 'Platelet count',
#     'Potassium': 'Potassium',
#     'RDW': 'RDW',
#     'RedBloodCells': 'Red blood cells',
#     'Sodium': 'Sodium',
#     'UreaNitrogen': 'Urea nitrogen',
#     'WhiteBloodCells': 'White blood cells',
#     'direct emrgency flag': 'Direct emergency',
#     'gender': 'Sex',
#     'last less than diff': 'Recent admission',
#     'night admission': 'Night admission',
#     'standardized age': 'Standardized age',
# }
#
# beta_units = {
#     'Admissions number 2': '2',
#     'Admissions number 3+': '3+',
#     'Anion gap': 'Abnormal',
#     'Bicarbonate': 'Abnormal',
#     'Calcium total': 'Abnormal',
#     'Chloride': 'Abnormal',
#     'Creatinine': 'Abnormal',
#     'Ethnicity black': 'Black',
#     'Ethnicity hispanic': 'Hispanic',
#     'Ethnicity other': 'Other',
#     'Ethnicity white': 'White',
#     'Glucose': 'Abnormal',
#     'Hematocrit': 'Abnormal',
#     'Hemoglobin': 'Abnormal',
#     'Insurance medicare': 'Medicare',
#     'Insurance other': 'Other',
#     'MCH': 'Abnormal',
#     'MCHC': 'Abnormal',
#     'MCV': 'Abnormal',
#     'Magnesium': 'Abnormal',
#     'Marital married': 'Married',
#     'Marital single': 'Single',
#     'Marital widowed': 'Widowed',
#     'Phosphate': 'Abnormal',
#     'Platelet count': 'Abnormal',
#     'Potassium': 'Abnormal',
#     'RDW': 'Abnormal',
#     'Red blood cells': 'Abnormal',
#     'Sodium': 'Abnormal',
#     'Urea nitrogen': 'Abnormal',
#     'White blood cells': 'Abnormal',
#     'Direct emergency': 'Yes',
#     'Sex': 'Female',
#     'Recent admission': 'Yes',
#     'Night admission': 'Yes',
#     'Standardized age': '',
# }


__all__ = ["TwoStagesCV", "TwoStagesCVExact", "PenaltyGridSearchCV", "PenaltyGridSearchCVExact"]


import pandas as pd
import numpy as np
from .fitters import TwoStagesFitter, TwoStagesFitterExact
import warnings
from copy import deepcopy
from sklearn.model_selection import KFold
pd.set_option("display.max_rows", 500)
warnings.filterwarnings('ignore')
slicer = pd.IndexSlice
from typing import Optional, List, Union
import psutil
from .evaluation import events_brier_score_at_t, events_integrated_brier_score, global_brier_score, \
    events_integrated_auc, global_auc, events_auc_at_t
from .model_selection import PenaltyGridSearch, PenaltyGridSearchExact
from time import time

WORKERS = psutil.cpu_count(logical=False)


class BaseTwoStagesCV(object):
    """
    This class implements K-fold cross-validation using TwoStagesFitters and TwoStagesFittersExact
    """

    def __init__(self):
        self.models = {}
        self.test_pids = {}
        self.results = pd.DataFrame()
        self.global_auc = {}
        self.integrated_auc = {}
        self.global_bs = {}
        self.integrated_bs = {}
        self.TwoStagesFitter_type = 'CoxPHFitter'

    def cross_validate(self,
                       full_df: pd.DataFrame,
                       n_splits: int = 5,
                       shuffle: bool = True,
                       seed: Union[int, None] = None,
                       fit_beta_kwargs: dict = {},
                       covariates=None,
                       event_type_col: str = 'J',
                       duration_col: str = 'X',
                       pid_col: str = 'pid',
                       x0: Union[np.array, int] = 0,
                       verbose: int = 2,
                       nb_workers: int = WORKERS,
                       metrics=['BS', 'IBS', 'GBS', 'AUC', 'IAUC', 'GAUC']):

        """
        This method implements K-fold cross-validation using TwoStagesFitters and full_df data.

        Args:
            full_df (pd.DataFrame): Data to cross validate.
            n_splits (int): Number of folds, defaults to 5.
            shuffle (bool): Shuffle samples before splitting to folds. Defaults to True.
            seed: Pseudo-random seed to KFold instance. Defaults to None.
            fit_beta_kwargs (dict, Optional): Keyword arguments to pass on to the estimation procedure. If different model for beta is desired, it can be defined here.
            covariates (list): list of covariates to be used in estimating the regression coefficients.
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in full_df).
            pid_col (str): Sample ID column name (must be a column in full_df).
            x0 (Union[numpy.array, int], Optional): initial guess to pass to scipy.optimize.minimize function
            verbose (int, Optional): The verbosity level of pandaallel
            nb_workers (int, Optional): The number of workers to pandaallel. If not sepcified, defaults to the number of workers available.
            metrics (str, list): Evaluation metrics.

        Returns:
            Results (pd.DataFrame): Cross validation metrics results
        """

        if isinstance(metrics, str):
            metrics = [metrics]

        self.models = {}
        self.kfold_cv = KFold(n_splits=n_splits, shuffle=shuffle, random_state=seed)

        if 'C' in full_df.columns:
            full_df = full_df.drop(['C'], axis=1)
        if 'T' in full_df.columns:
            full_df = full_df.drop(['T'], axis=1)

        for i_fold, (train_index, test_index) in enumerate(self.kfold_cv.split(full_df)):
            self.test_pids[i_fold] = full_df.iloc[test_index][pid_col].values
            train_df, test_df = full_df.iloc[train_index], full_df.iloc[test_index]
            if self.TwoStagesFitter_type == 'Exact':
                fold_fitter = TwoStagesFitterExact()
            else:
                fold_fitter = TwoStagesFitter()
            print(f'Fitting fold {i_fold+1}/{n_splits}')
            fold_fitter.fit(df=train_df,
                            covariates=covariates,
                            event_type_col=event_type_col,
                            duration_col=duration_col,
                            pid_col=pid_col,
                            x0=x0,
                            fit_beta_kwargs=fit_beta_kwargs,
                            verbose=verbose,
                            nb_workers=nb_workers)

            #self.models[i_fold] = deepcopy(fold_fitter)
            self.models[i_fold] = fold_fitter

            pred_df = self.models[i_fold].predict_prob_events(test_df)

            for metric in metrics:
                if metric == 'IAUC':
                    self.integrated_auc[i_fold] = events_integrated_auc(pred_df, event_type_col=event_type_col,
                                                                        duration_col=duration_col)
                elif metric == 'GAUC':
                    self.global_auc[i_fold] = global_auc(pred_df, event_type_col=event_type_col,
                                                                  duration_col=duration_col)
                elif metric == 'IBS':
                    self.integrated_bs[i_fold] = events_integrated_brier_score(pred_df, event_type_col=event_type_col,
                                                                                        duration_col=duration_col)
                elif metric == 'GBS':
                    self.global_bs[i_fold] = global_brier_score(pred_df, event_type_col=event_type_col,
                                                                         duration_col=duration_col)
                elif metric == 'AUC':
                    tmp_res = events_auc_at_t(pred_df, event_type_col=event_type_col,
                                                       duration_col=duration_col)
                    tmp_res = pd.concat([tmp_res], keys=[i_fold], names=['fold'])
                    tmp_res = pd.concat([tmp_res], keys=[metric], names=['metric'])
                    self.results = pd.concat([self.results, tmp_res], axis=0)
                elif metric == 'BS':
                    tmp_res = events_brier_score_at_t(pred_df, event_type_col=event_type_col,
                                                               duration_col=duration_col)
                    tmp_res = pd.concat([tmp_res], keys=[i_fold], names=['fold'])
                    tmp_res = pd.concat([tmp_res], keys=[metric], names=['metric'])
                    self.results = pd.concat([self.results, tmp_res], axis=0)

        return self.results


class BasePenaltyGridSearchCV(object):
    """
    This class implements K-fold cross-validation of the PenaltyGridSearch
    """

    def __init__(self):
        self.folds_grids = {}
        self.test_pids = {}
        self.global_auc = {}
        self.integrated_auc = {}
        self.global_bs = {}
        self.integrated_bs = {}
        self.TwoStagesFitter_type = 'CoxPHFitter'

    def cross_validate(self,
                       full_df: pd.DataFrame,
                       l1_ratio: float,
                       penalizers: list,
                       n_splits: int = 5,
                       shuffle: bool = True,
                       seed: Union[int, None] = None,
                       event_type_col: str = 'J',
                       duration_col: str = 'X',
                       pid_col: str = 'pid',
                       twostages_fit_kwargs: dict = {'nb_workers': WORKERS},
                       metrics=['IBS', 'GBS', 'IAUC', 'GAUC']) -> pd.DataFrame:

        """
        This method implements K-fold cross-validation using PenaltyGridSearch and full_df data.

        Args:
            full_df (pd.DataFrame): Data to cross validate.
            l1_ratio (float): regularization ratio for the CoxPHFitter (see lifelines.fitters.coxph_fitter.CoxPHFitter documentation).
            penalizers (list): penalizer options for each event (see lifelines.fitters.coxph_fitter.CoxPHFitter documentation).
            n_splits (int): Number of folds, defaults to 5.
            shuffle (boolean): Shuffle samples before splitting to folds. Defaults to True.
            seed: Pseudo-random seed to KFold instance. Defaults to None.
            event_type_col (str): The event type column name (must be a column in df), Right-censored sample (i) is indicated by event value 0, df.loc[i, event_type_col] = 0.
            duration_col (str): Last follow up time column name (must be a column in full_df).
            pid_col (str): Sample ID column name (must be a column in full_df).
            twostages_fit_kwargs (dict): keyword arguments to pass to each TwoStagesFitter.
            metrics (str, list): Evaluation metrics.

        Returns:
            gauc_output_df (pd.DataFrame): Global AUC k-fold mean and standard error for all possible combination of the penalizers.
        """

        if isinstance(metrics, str):
            metrics = [metrics]

        self.folds_grids = {}
        self.kfold_cv = KFold(n_splits=n_splits, shuffle=shuffle, random_state=seed)

        if 'C' in full_df.columns:
            full_df = full_df.drop(['C'], axis=1)
        if 'T' in full_df.columns:
            full_df = full_df.drop(['T'], axis=1)

        for i_fold, (train_index, test_index) in enumerate(self.kfold_cv.split(full_df)):
            print(f'Starting fold {i_fold+1}/{n_splits}')
            start = time()
            self.test_pids[i_fold] = full_df.iloc[test_index][pid_col].values
            train_df, test_df = full_df.iloc[train_index], full_df.iloc[test_index]
            if self.TwoStagesFitter_type == 'Exact':
                fold_pgs = PenaltyGridSearchExact()
            else:
                fold_pgs = PenaltyGridSearch()

            fold_pgs.evaluate(train_df=train_df,
                              test_df=test_df,
                              l1_ratio=l1_ratio,
                              penalizers=penalizers,
                              metrics=metrics,
                              seed=seed,
                              event_type_col=event_type_col,
                              duration_col=duration_col,
                              pid_col=pid_col,
                              twostages_fit_kwargs=twostages_fit_kwargs)

            self.folds_grids[i_fold] = fold_pgs

            for metric in metrics:
                if metric == 'GAUC':
                    self.global_auc[i_fold] = fold_pgs.convert_results_dict_to_df(fold_pgs.global_auc)
                elif metric == 'IAUC':
                    self.integrated_auc[i_fold] = fold_pgs.convert_results_dict_to_df(fold_pgs.integrated_auc)
                elif metric == 'GBS':
                    self.global_bs[i_fold] = fold_pgs.convert_results_dict_to_df(fold_pgs.global_bs)
                elif metric == 'IBS':
                    self.integrated_bs[i_fold] = fold_pgs.convert_results_dict_to_df(fold_pgs.integrated_bs)

            end = time()
            print(f'Finished fold {i_fold+1}/{n_splits}, {int(end-start)} seconds')

        if 'GAUC' in metrics:
            res = [v for k, v in self.global_auc.items()]
            gauc_output_df = pd.concat([pd.concat(res, axis=1).mean(axis=1),
                                        pd.concat(res, axis=1).std(axis=1)],
                                       keys=['Mean', 'SE'], axis=1)
        else:
            gauc_output_df = pd.DataFrame()
        return gauc_output_df


class PenaltyGridSearchCV(BasePenaltyGridSearchCV):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'CoxPHFitter'


class PenaltyGridSearchCVExact(BasePenaltyGridSearchCV):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'Exact'


class TwoStagesCV(BaseTwoStagesCV):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'CoxPHFitter'


class TwoStagesCVExact(BaseTwoStagesCV):

    def __init__(self):
        super().__init__()
        self.TwoStagesFitter_type = 'Exact'



import numpy as np
from pydts.examples_utils.simulations_data_config import *
from pydts.config import *
import pandas as pd
from scipy.special import expit
from pandarallel import pandarallel


def sample_los(new_patient, age_mean, age_std, bmi_mean, bmi_std, coefs=COEFS, baseline_hazard_scale=8,
               los_bounds=[1, 150]):
    # Columns normalization:
    new_patient[AGE_COL] = (new_patient[AGE_COL] - age_mean) / age_std
    new_patient[GENDER_COL] = 2 * (new_patient[GENDER_COL] - 0.5)
    new_patient[BMI_COL] = (new_patient[BMI_COL] - bmi_mean) / bmi_std
    new_patient[SMOKING_COL] = new_patient[SMOKING_COL] - 1
    new_patient[HYPERTENSION_COL] = 2 * (new_patient[HYPERTENSION_COL] - 0.5)
    new_patient[DIABETES_COL] = 2 * (new_patient[DIABETES_COL] - 0.5)
    new_patient[ART_FIB_COL] = 2 * (new_patient[ART_FIB_COL] - 0.5)
    new_patient[COPD_COL] = 2 * (new_patient[COPD_COL] - 0.5)
    new_patient[CRF_COL] = 2 * (new_patient[CRF_COL] - 0.5)
    new_patient = pd.Series(new_patient)

    # Baseline hazard
    baseline_hazard = np.random.exponential(scale=baseline_hazard_scale)

    # Patient's correction
    beta_x = coefs.dot(new_patient[coefs.index])

    # Sample, round (for ties), and clip to bounds the patient's length of stay at the hospital
    los = np.clip(np.round(baseline_hazard * np.exp(beta_x)), a_min=los_bounds[0], a_max=los_bounds[1])
    los_death = np.nan if new_patient[IN_HOSPITAL_DEATH_COL] == 0 else los
    return los, los_death


def hide_weight_info(row):
    #admyear = row[ADMISSION_YEAR_COL]
    p_weight = 0.3 #+ int(admyear > (min_year + 3)) * 0.8 * ((admyear - min_year) / (max_year - min_year))
    sample_weight = np.random.binomial(1, p=p_weight)
    if sample_weight == 0:
        row[WEIGHT_COL] = np.nan
        row[BMI_COL] = np.nan
    return row


def main(seed=0, N_patients=DEFAULT_N_PATIENTS, output_dir=OUTPUT_DIR, filename=SIMULATED_DATA_FILENAME):
    # Set random seed for consistent sampling
    np.random.seed(seed)

    # Female - 1, Male - 0
    gender = np.random.binomial(n=1, p=0.5, size=N_patients)

    simulated_patients_df = pd.DataFrame()

    for p in range(N_patients):
        # Sample gender dependent age for each patient
        age = np.round(np.random.normal(loc=72 + 5 * gender[p], scale=12), decimals=1)

        # Random sample admission year
        admyear = np.random.randint(low=min_year, high=max_year)

        # Sample gender dependent height
        height = np.random.normal(loc=175 - 5 * gender[p], scale=7)

        # Sample height, gender and age dependent weight
        weight = np.random.normal(loc=(height / 175) * 80 - 5 * gender[p] + (age / 20), scale=8)

        # Calculate body mass index (BMI) from weight and height
        bmi = weight / ((height / 100) ** 2)

        # Random sample of previous admissions
        admserial = np.clip(np.round(np.random.lognormal(mean=0, sigma=0.75)), 1, 20)

        # Random sample of categorical smoking status: No - 0, Previously - 1, Currently - 2
        smoking = np.random.choice([0, 1, 2], p=[0.5, 0.3, 0.2])

        # Sample patient's preconditions based on gender, age, BMI, and smoking status with limits on the value of p
        pre_p = np.clip((bmi_coef * bmi + gender_coef * gender[p] + age_coef * age + smk_coef * smoking),
                        a_min=0.05, a_max=max_p)
        hypertension = np.random.binomial(n=1, p=pre_p)
        diabetes = np.random.binomial(n=1, p=pre_p + bmi_coef * bmi)
        artfib = np.random.binomial(n=1, p=pre_p)  # Arterial Fibrillation
        copd = np.random.binomial(n=1, p=pre_p + smk_coef * smoking)  # Chronic Obstructive Pulmonary Disease
        crf = np.random.binomial(n=1, p=pre_p)  # Chronic Renal Failure

        new_patient = {
            PATIENT_NO_COL: p,
            AGE_COL: age,
            GENDER_COL: gender[p],
            ADMISSION_YEAR_COL: int(admyear),
            FIRST_ADMISSION_COL: int(admserial == 1),
            ADMISSION_SERIAL_COL: int(admserial),
            WEIGHT_COL: weight,
            HEIGHT_COL: height,
            BMI_COL: bmi,
            SMOKING_COL: smoking,
            HYPERTENSION_COL: hypertension,
            DIABETES_COL: diabetes,
            ART_FIB_COL: artfib,
            COPD_COL: copd,
            CRF_COL: crf,
        }

        simulated_patients_df = simulated_patients_df.append(new_patient, ignore_index=True)

    # [age, gender, ADMISSION_YEAR, FIRST_ADMISSION, ADMISSION_SERIAL, WEIGHT_COL, HEIGHT_COL, BMI, SMOKING_COL, 5*preconditions]
    #b1 = [-0.003, 0.05, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    b2 = [0.003, -0.05, 0, 0, 0, 0.003, 0.0001, 0.005, 0.05, 0.02, 0.02, 0.02, 0.02, 0.02]

    real_coef_dict = {
        "alpha": {
            1: lambda t: -2.2 - 0.2 * np.log(t),
            2: lambda t: -2.3 - 0.2 * np.log(t)
        },
        "beta": {
            1: [-1*b for b in b2],
            2: [b for b in b2]
        }
    }

    # Sample event type and relative time
    events_df = new_sample_logic(simulated_patients_df.set_index(PATIENT_NO_COL), j_events=2, d_times=30,
                                             real_coef_dict=real_coef_dict)
    simulated_patients_df = pd.concat([simulated_patients_df, events_df], axis=1)

    simulated_patients_df[DEATH_RELATIVE_COL] = np.nan
    simulated_patients_df.loc[simulated_patients_df.J == 1, DEATH_RELATIVE_COL] = simulated_patients_df.loc[
        simulated_patients_df.J == 1, 'T'].values
    simulated_patients_df = simulated_patients_df.rename(columns={'T': DISCHARGE_RELATIVE_COL})
    simulated_patients_df.loc[simulated_patients_df.J == 0, DISCHARGE_RELATIVE_COL] = 31
    simulated_patients_df = simulated_patients_df.drop('J', axis=1)

    # Remove weight and bmi based on admission year
    simulated_patients_df = simulated_patients_df.apply(hide_weight_info, axis=1)

    simulated_patients_df[DEATH_MISSING_COL] = simulated_patients_df[DEATH_RELATIVE_COL].isnull().astype(int)
    simulated_patients_df[IN_HOSPITAL_DEATH_COL] = simulated_patients_df[DEATH_RELATIVE_COL].notnull().astype(int)
    simulated_patients_df[RETURNING_PATIENT_COL] = pd.cut(simulated_patients_df[ADMISSION_SERIAL_COL],
                                                        bins=ADMISSION_SERIAL_BINS, labels=ADMISSION_SERIAL_LABELS)

    simulated_patients_df.set_index(PATIENT_NO_COL).to_csv(os.path.join(output_dir, filename))


def default_sampling_logic(Z, d_times):
    alpha1t = -1 -0.3*np.log(np.arange(start=1, stop=d_times+1))
    beta1 = -np.log([0.8, 3, 3, 2.5, 2])
    alpha2t = -1.75 -0.15*np.log(np.arange(start=1, stop=d_times+1))
    beta2 = -np.log([1, 3, 4, 3, 2])
    hazard1 = expit(alpha1t+(Z*beta1).sum())
    hazard2 = expit(alpha2t+(Z*beta2).sum())
    surv_func = np.array([1, *np.cumprod(1-hazard1-hazard2)[:-1]])
    proba1 = hazard1*surv_func
    proba2 = hazard2*surv_func
    sum1 = np.sum(proba1)
    sum2 = np.sum(proba2)
    probj1t = proba1 / sum1
    probj2t = proba2 / sum2
    j_i = np.random.choice(a=[0, 1, 2], p=[1-sum1-sum2, sum1, sum2])
    if j_i == 0:
        T_i = d_times
    elif j_i == 1:
        T_i = np.random.choice(a=np.arange(1, d_times+1), p=probj1t)
    else:
        T_i = np.random.choice(a=np.arange(1, d_times+1), p=probj2t)
    return j_i, T_i


def calculate_jt(sums, probs_jt, d_times, j_events):
    events = range(1, j_events + 1)
    temp_sums = pd.concat([1 - sum(sums), *sums], axis=1, keys=[0, *events])

    j_df = (temp_sums.cumsum(1) > np.random.rand(temp_sums.shape[0])[:, None]).idxmax(axis=1).to_frame('J')

    temp_ts = []
    for j in events:
        rel_j = j_df.query("J==@j").index
        prob_df = probs_jt[j - 1]  # the prob j to sample from
        # sample T
        temp_ts.append((prob_df.loc[rel_j].cumsum(1) >= np.random.rand(rel_j.shape[0])[:, None]).idxmax(axis=1))

    temp_ts.append(pd.Series(d_times+1, index=j_df.query('J==0').index))

    j_df["T"] = pd.concat(temp_ts).sort_index()
    return j_df


def new_sample_logic(patients_df: pd.DataFrame, j_events: int, d_times: int, real_coef_dict: dict) -> pd.DataFrame:
    """
    A quicker sample logic, that uses coefficients supplied by the user
    """
    events = range(1, j_events + 1)
    a_t = {event: {t: real_coef_dict['alpha'][event](t) for t in range(1, d_times+1)} for event in events}
    b = pd.concat([patients_df.dot(real_coef_dict['beta'][j]) for j in events], axis=1, keys=events)

    hazards = [pd.concat([expit(a_t[j][t] + b[j]) for t in range(1, d_times + 1)],
                         axis=1, keys=(range(1, d_times + 1))) for j in events]
    surv_func = pd.concat([pd.Series(1, index=hazards[0].index),
                           (1 - sum(hazards)).cumprod(axis=1).iloc[:, :-1]], axis=1)
    surv_func.columns += 1

    probs = [hazard * surv_func for hazard in hazards]
    sums = [prob.sum(axis=1) for prob in probs]
    probjt = [prob.div(sumj, axis=0) for prob, sumj in zip(probs, sums)]

    ret = calculate_jt(sums, probjt, d_times, j_events)
    return ret


def generate_quick_start_df(n_patients=10000, d_times=30, j_events=2, n_cov=5, seed=0, pid_col='pid',
                            real_coef_dict: dict = DEFAULT_REAL_COEF_DICT, sampling_logic=new_sample_logic,
                            censoring_prob=1.):
    np.random.seed(seed)
    assert real_coef_dict is not None, "The user should supply the coefficients of the experiment"
    covariates = [f'Z{i + 1}' for i in range(n_cov)]
    patients_df = pd.DataFrame(data=np.random.uniform(low=0.0, high=1.0, size=[n_patients, n_cov]),
                               columns=covariates)
    sampled = sampling_logic(patients_df, j_events, d_times, real_coef_dict)
    patients_df = pd.concat([patients_df, sampled], axis=1)
    patients_df.index.name = pid_col
    patients_df['C'] = np.where(np.random.rand(n_patients) < censoring_prob,
                                np.random.randint(low=1, high=d_times+1,
                                                  size=n_patients), d_times+1)
    patients_df['X'] = patients_df[['T', 'C']].min(axis=1)
    patients_df.loc[patients_df['C'] < patients_df['T'], 'J'] = 0
    return patients_df.reset_index()


if __name__ == "__main__":
    main(N_patients=50000)
    #generate_quick_start_df(n_patients=2)
